#include "NVSExplorer.hpp"

// C++ system level
#include <set>
#include <cmath>
#include <vector>
#include <cstring>     // memset, memcpy, strncmp
// #include <cstdlib>
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_err.h"
// #include "driver/uart.h"

// project specific
#include <SpiderLib/Util.hpp>
#include "SpiderLib/NumberTypes.hpp"

// qthing stuff
#include <qthing>
#include <qthing/mqtt_common.hpp>




namespace SiliconTorch {

  namespace NVSExplorer {


    const char* TAG = "NVSExplorer";


    NVSExplorer::NVSExplorer() {

      ESP_LOGI(TAG, "Starting NVSExplorer[📡]");

      nvs_flash_init();



      u16 deviceTopicLEN = requestTopic("").length();

      auto wrapMessageHandler = [deviceTopicLEN](std::function<void(std::vector<std::string>&, qthing::multipart_message_t)> handler) {
        return [handler, deviceTopicLEN](qthing::multipart_message_t msg) {
          if (msg.length == msg.total_length && msg.offset == 0) {

            std::string localTopic(msg.topic.c_str() + deviceTopicLEN);
            std::vector<std::string> topicParts = SpiderLib::Util::split(localTopic, '/');

            handler(topicParts, msg);
          } else {
            ESP_LOGW(TAG, "Multipart messages are not supported atm.  topic[ %s ]", msg.topic.c_str());
          }
        };
      };



      auto lsKeys = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto keys = listKeys(topic[1]);
        std::string msgOut = SpiderLib::Util::join(keys, delimiter);

        qthing::publish_message(responseTopic(topic[1]), msgOut);

        if (buffer != NULL) delete buffer;
      };

      qthing::add_binary_message_callback(requestTopic("ls/+"), wrapMessageHandler(lsKeys));



      auto lsNS = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto nameSpaces = listNamespaces();
        std::string msgOut = SpiderLib::Util::join(nameSpaces, delimiter);

        qthing::publish_message(responseTopic("ls"), msgOut);

        if (buffer != NULL) delete buffer;
      };

      qthing::add_binary_message_callback(requestTopic("ls"), wrapMessageHandler(lsNS));



      auto rmKey = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        this->rmKey(topic[1], topic[2]);
      };

      qthing::add_binary_message_callback(requestTopic("rm/+/+"), wrapMessageHandler(rmKey));



      auto getFloat = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        float value = this->getFloat(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_float", topic[1], topic[2]), std::string(buffer));
        } else {
          // publish error…?
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_float/+/+"), wrapMessageHandler(getFloat));



      auto getDouble = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        double value = this->getDouble(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_double", topic[1], topic[2]), std::string(buffer));
        } else {
          // publish error…?
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_double/+/+"), wrapMessageHandler(getDouble));


    }


    float NVSExplorer::getFloat(const std::string& nameSpace, const std::string& key) {

      float value = std::nanf("1337");
      u32* valptr = reinterpret_cast<u32*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u32(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }

    double NVSExplorer::getDouble(const std::string& nameSpace, const std::string& key) {

      double value = std::nan("1337");
      u64* valptr = reinterpret_cast<u64*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u64(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }


    bool NVSExplorer::setFloat(const std::string& nameSpace, const std::string& key, float value) {

      bool success    = true;
      u32 intval = reinterpret_cast<u32*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u32(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setDouble(const std::string& nameSpace, const std::string& key, double value) {

      bool success    = true;
      u64 intval = reinterpret_cast<u64*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u64(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }


    void NVSExplorer::rmKey(const std::string& nameSpace, const std::string& key) {
      nvs_handle_t _nvs;
      ESP_LOGW(TAG, "rmKey() -> %s / %s", nameSpace.c_str(), key.c_str());
      ESP_LOGW(TAG, "open() -> %s", esp_err_to_name( nvs_open(nameSpace.c_str(), NVS_READWRITE, &_nvs) ));
      ESP_LOGW(TAG, "erase() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
      ESP_LOGW(TAG, "commit() -> %s", esp_err_to_name( nvs_commit(_nvs) ));
      ESP_LOGW(TAG, "erase²() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
    }

    std::set<std::string> NVSExplorer::listKeys(const std::string& nameSpace) {

      std::set<std::string> keys;

      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

        keys.insert(info.key);
      }

      return keys;
    }

    std::set<std::string> NVSExplorer::listNamespaces() {

      std::set<std::string> namespaces;

      nvs_iterator_t it = nvs_entry_find("nvs", NULL, NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

        namespaces.insert(info.namespace_name);
      }

      return namespaces;
    }


    i64  getSignedInt(const std::string& nameSpace, const std::string& key, i64 defaultValue) {

      // Lets first test & evaluate the other function before duplication………

      return -1337LL;
    }


    u64 getUnsignedInt(const std::string& nameSpace, const std::string& key, u64 defaultValue) {

      u64 value = defaultValue;
      nvs_type_t entryType = NVS_TYPE_ANY;

      // search for the key to get its type
      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t entry;
        nvs_entry_info(it, &entry);

        if (key.compare(entry.key) == 0) {
          entryType = entry.type;
          nvs_release_iterator(it);
          break;
        }

        it = nvs_entry_next(it);
      }

      if (entryType == NVS_TYPE_ANY)  // key not found
        return defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) != ESP_OK )  // namespace not found
        return defaultValue;

      u64* valptr = &value;
      switch (entryType) {
        case NVS_TYPE_U8:  { nvs_get_u8(nvs,  key.c_str(), ( u8*)valptr); break; }
        case NVS_TYPE_I8:  { nvs_get_i8(nvs,  key.c_str(), (  i8*)valptr); break; }
        case NVS_TYPE_U16: { nvs_get_u16(nvs, key.c_str(), (u16*)valptr); break; }
        case NVS_TYPE_I16: { nvs_get_i16(nvs, key.c_str(), ( i16*)valptr); break; }
        case NVS_TYPE_U32: { nvs_get_u32(nvs, key.c_str(), (u32*)valptr); break; }
        case NVS_TYPE_I32: { nvs_get_i32(nvs, key.c_str(), ( i32*)valptr); break; }
        case NVS_TYPE_U64: { nvs_get_u64(nvs, key.c_str(), (u64*)valptr); break; }
        case NVS_TYPE_I64: { nvs_get_i64(nvs, key.c_str(), ( i64*)valptr); break; }
        default: {}
      }

      nvs_close(nvs);
      return value;
    }



    std::string NVSExplorer::deviceTopic(const std::string& suffix) {
      return std::string(DEVICE_NAMESPACE + "NVS/") + suffix;
    }


    std::string NVSExplorer::requestTopic(const std::string& cmd) {
      return deviceTopic( std::string("rq/") + cmd );
    }

    std::string NVSExplorer::requestTopic(const std::string& cmd, const std::string& nameSpace) {
      return deviceTopic( std::string("rq/") + cmd + std::string("/") + nameSpace );
    }

    std::string NVSExplorer::requestTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key) {
      return deviceTopic( std::string("rq/") + cmd + std::string("/") + nameSpace + std::string("/") + key );
    }


    std::string NVSExplorer::responseTopic(const std::string& cmd) {
      return deviceTopic( std::string("rs/") + cmd );
    }

    std::string NVSExplorer::responseTopic(const std::string& cmd, const std::string& nameSpace) {
      return deviceTopic( std::string("rs/") + cmd + std::string("/") + nameSpace );
    }

    std::string NVSExplorer::responseTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key) {
      return deviceTopic( std::string("rs/") + cmd + std::string("/") + nameSpace + std::string("/") + key );
    }

  }
}
